#include "stdafx.h"
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "network.h"

#define SIG_HANDLE(SIG_NAME, SIG_FUNC) glade_xml_signal_connect(xml, SIG_NAME, (GCallback) SIG_FUNC)
#define ST_ON (status == STS_ON)
#define NEW_CHAR(x) ((char*)malloc(sizeof(char)*x))

// Glade Data
GladeXML* xml; 
GtkWidget*  main_window;
GtkTextBuffer* log_area_buffer;
GtkWidget* dialog;
GtkWidget* info_dialog;
// Glade Data

// App data
typedef enum {
	MSG_NORMAL, MSG_ATTENTION, MSG_ERROR
} MSG_TYPE;
typedef enum {
	STS_INV, STS_OFF, STS_ON
} MD_STATE;
PROCESS_INFORMATION pi;
MD_STATE status;
int full_screen = 1;
FILE* conf;
FILE* clog;
char* net_msgs;
int net_msg_cnt = -1;
HANDLE net_thread;
// App data

void onZoomSpinValueChanged();
void onPanSpinValueChanged();
void errorMessage(char* );
void guiSendNetwork(char* buf);

// Configuration utility functions
void initConfigure(){
	// Check if the file exist, if not make one
	conf = fopen("conf", "r");
	if(!conf){
		conf = fopen("conf", "w");
		fprintf(conf, "# Configuration file: max line length is 512 characters unless specified\n");
		fprintf(conf, "# Location of the MPIRun utility\n");
		fprintf(conf, "MPI:C:\\Program Files\\MPICH\\mpd\\bin\\MPIRun.exe\n");
		fprintf(conf, "# Location of the multidisplay program\n");
		fprintf(conf, "CMD:C:\\Program Files\\CECA MPI_OpenGL2\\display\\MPI_OpenGL2.exe\n");
		fprintf(conf, "# Grid size (width x height) max of 16x9\n");
		fprintf(conf, "GRI:1 1\n");
		fprintf(conf, "# Full Screen (1 = Full Screen, 0 = No Full Screen)\n");
		fprintf(conf, "FLS:0\n");
		fprintf(conf, "# Image file \n");
		fprintf(conf, "IMG:C:\\Documents and Settings\\All Users\\Documents\\My Pictures\\Sample Pictures\\Sunset.jpg\n");
        fprintf(conf, "# Zoom Factor \n");
		fprintf(conf, "ZOM: 5\n");
        fprintf(conf, "# Pan Factor \n");
		fprintf(conf, "PAN: 5\n");
		fclose(conf);
		conf = fopen("conf", "r");
	}
	// Read configurations
	char* conf_line = NEW_CHAR(512);
	char* mpi_loc = NEW_CHAR(512);
	char* md_loc = NEW_CHAR(512);
	char* img_loc = NEW_CHAR(512);
	ZeroMemory(conf_line, 512);
	ZeroMemory(mpi_loc, 512);
	ZeroMemory(md_loc, 512);
	ZeroMemory(img_loc, 512);
	int g_width = 1;
	int g_height = 1;
	int full_scr = 1;
	while(fgets(conf_line, 512, conf)){
		conf_line[strlen(conf_line) - 1] = 0;
		switch(conf_line[0]){
		case 'M':{
			strcpy(mpi_loc, &conf_line[4]); break;
				 }
		case 'C':{
			strcpy(md_loc, &conf_line[4]); break;
				 }
		case 'G':{
			sscanf(&conf_line[4], "%ld %ld", &g_width, &g_height); break;
				 }
		case 'F':{
			sscanf(&conf_line[4], "%ld", &full_scr); break;
				 }
		case 'I':{
			strcpy(img_loc, &conf_line[4]); break;
				 }
        case 'Z':{
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "zoom_spin")), atof(&conf_line[4])); break;
				 }
        case 'P':{
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "pan_spin")), atof(&conf_line[4])); break;
				 }
		default:{
			// nothing...
				 }
		}
	}
	fclose(conf);
	// Apply configurations
	if(strlen(mpi_loc) > 1)
		gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(xml, "mpirun_path_text")), mpi_loc);
	if(strlen(md_loc) > 1)
		gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(xml, "md_path_text")), md_loc);
	if(strlen(img_loc) > 1)
		gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(xml, "image_file_path")), img_loc);
	if(g_width > 0 && g_width < 6) 
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "grid_width_spin")), (gdouble) g_width);
	if(g_height > 0 && g_height < 6) 
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "grid_height_spin")), (gdouble) g_height);
	if(full_scr == 1) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "full_screen_checkbutton")), TRUE);
		full_screen = 0;
	}else{
		full_screen = 1;
	}
    free(conf_line);
	free(mpi_loc);
	free(md_loc);
	free(img_loc);
}

void finiConfigure(){
	conf = fopen("conf", "w");
	if(!conf){
		errorMessage("Couldn't save configuration file");
	}else{
		int width = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "grid_width_spin")));
		int height = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "grid_height_spin")));		
		fprintf(conf, "# Saved on %ld\n", time(NULL));
		fprintf(conf, "# Configuration file: max line length is 512 characters unless specified\n");
		fprintf(conf, "# Location of the MPIRun utility\n");
		fprintf(conf, "MPI:%s\n", gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(xml, "mpirun_path_text"))));
		fprintf(conf, "# Location of the multidisplay program\n");
		fprintf(conf, "CMD:%s\n", gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(xml, "md_path_text"))));
		fprintf(conf, "# Grid size (width x height) max of 5x5\n");
		fprintf(conf, "GRI:%d %d\n", width, height);
		fprintf(conf, "# Full Screen (1 = Full Screen, 0 = No Full Screen)\n");
		fprintf(conf, "FLS:%d\n", ((full_screen==1)? 0 : 1));
		fprintf(conf, "# Image file \n");
		fprintf(conf, "IMG:%s\n", gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(xml, "image_file_path"))));
        fprintf(conf, "# Zoom Factor \n");
		fprintf(conf, "ZOM: %f\n", gtk_spin_button_get_value(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "zoom_spin"))));
        fprintf(conf, "# Pan Factor \n");
		fprintf(conf, "PAN: %f\n", gtk_spin_button_get_value(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "pan_spin"))));
		fclose(conf);
	}
}
// Configuration utility functions


void log(char *log, MSG_TYPE type){	
	GtkTextIter iter;
	GtkTextTag *tag = NULL;
	gtk_text_buffer_get_end_iter(log_area_buffer, &iter);

	switch(type){
	case MSG_NORMAL:{
		tag = gtk_text_buffer_create_tag (log_area_buffer, NULL, 
			"foreground", "black", 
			NULL);
		break;
					}
	case MSG_ATTENTION:{
		tag = gtk_text_buffer_create_tag (log_area_buffer, NULL, 
			"foreground", "orange", 
			NULL);
		break;
					 }
	case MSG_ERROR:{
		tag = gtk_text_buffer_create_tag (log_area_buffer, NULL, 
			"foreground", "red", 
			NULL);
		break;
				   }
	default:{
	tag = gtk_text_buffer_create_tag (log_area_buffer, NULL, 
		"foreground", "brown", 
		NULL);
	break;
			   }
	}
	gtk_text_buffer_insert_with_tags (log_area_buffer, &iter, log, -1, tag, NULL);
	gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(glade_xml_get_widget(xml, "log")), &iter, 0, FALSE, 0, 0);
}

void selectFile(char* entry_name){
	GtkWidget* entry = glade_xml_get_widget(xml, entry_name);	
	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT){
		char* file_name = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
		char* log_string = NEW_CHAR(512);
		sprintf(log_string, "File '%s' selected\n", file_name);
		log(log_string, MSG_ATTENTION);
		gtk_entry_set_text(GTK_ENTRY(entry), file_name);
        free(log_string);
	}
    gtk_widget_hide (dialog);
}

void infoMessage(char* message){
	info_dialog = gtk_message_dialog_new (GTK_WINDOW(main_window),
		GTK_DIALOG_DESTROY_WITH_PARENT,
		GTK_MESSAGE_INFO,
		GTK_BUTTONS_OK,
		message);
	gtk_dialog_run (GTK_DIALOG (info_dialog));
	gtk_widget_destroy(info_dialog);
}

void errorMessage(char* message){
	info_dialog = gtk_message_dialog_new (GTK_WINDOW(main_window),
		GTK_DIALOG_DESTROY_WITH_PARENT,
		GTK_MESSAGE_ERROR,
		GTK_BUTTONS_OK,
		message);
	gtk_dialog_run (GTK_DIALOG (info_dialog));	
	gtk_widget_destroy(info_dialog);
}

int queryMessage(char* message){
	info_dialog = gtk_message_dialog_new (GTK_WINDOW(main_window),
		GTK_DIALOG_DESTROY_WITH_PARENT,
		GTK_MESSAGE_QUESTION,
		GTK_BUTTONS_OK_CANCEL,
		message);
	int response = gtk_dialog_run (GTK_DIALOG (info_dialog));
	gtk_widget_destroy(info_dialog);
	return response;
}

// Menu signals
void onQuit1Activate(){  
	if(ST_ON) {
		errorMessage("Please stop current session first");
		return;
	}
	int response = queryMessage("Exit Tiled Display?");	
	if(response == GTK_RESPONSE_OK) {
		finiConfigure();
		exit(0);
	}
}

void onHelp1Activate(){
	infoMessage("Please contact Stephanie-Smullen@utc.edu\nOr Mohamed-Ibrahim@utc.edu\nCECA Tiled Display Version 1.2 (March '05)");
}

// Menu signals


//  Executable paths signals
void onMpiBrowseClicked(){
	selectFile("mpirun_path_text");
}

void onMultdispClicked() {
	selectFile("md_path_text");
}
//  Executable paths signals


// Start Signals
void onStartClicked() {
	if(ST_ON){
		errorMessage("Please stop current session first");
		return;
	}
	const char* mpi_loc = gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(xml, "mpirun_path_text")));
	const  char* multi_loc = gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(xml, "md_path_text")));
	FILE* mpi_file = fopen(mpi_loc, "r");
	FILE* multi_file = fopen(multi_loc, "r");
	if(!mpi_file){
		errorMessage("Please specify a valid path to 'MPIRun'.\nSee the 'Configuration' tab");
		return;
	}
	fclose(mpi_file);
	if(!multi_file){
		errorMessage("Please specify a valid path to 'Tiled Display'.\nSee the 'Configuration' tab");
		return;
	}
	fclose(multi_file);

	int length_mpi = strlen(mpi_loc);
	int length_multi = strlen(multi_loc);
	if(length_mpi > 1 && length_multi > 1){
		int width = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "grid_width_spin")));
		int height = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "grid_height_spin")));		
		char* cmd_line = NEW_CHAR(1024);
		sprintf(cmd_line,"\"%s\" -np %d \"%s\" %d %d %d", 
			mpi_loc, (width*height)+1, 
			multi_loc, width, height, full_screen);
		STARTUPINFO si;				
		ZeroMemory( &si, sizeof(si) );
		si.cb = sizeof(si);
		ZeroMemory( &pi, sizeof(pi) );

		if(!CreateProcess(NULL,
			cmd_line, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &si, &pi))
			errorMessage("Failed to launch display clients!");	
		
		int good = initNetwork(9000);


		if(good == NETWORK_ERROR){
			status = STS_OFF;
			log("Couldn't connect to client!", MSG_ERROR);
			finiNetwork();
		}else{
			status = STS_ON;
		}
        onZoomSpinValueChanged();
        Sleep(1000);
        onPanSpinValueChanged();
        free(cmd_line);
	} else {
		errorMessage("Please specify the path to both 'MPIRun' and 'Tiled Display'.\nSee the 'Configuration' tab");
	}
}

void onStopClicked() {	
	if(ST_ON){
		guiSendNetwork("9999");
		status = STS_OFF;
		finiNetwork();
		LPDWORD lpExitCode = (LPDWORD)malloc(sizeof(unsigned long));
		GetExitCodeProcess(pi.hProcess, lpExitCode);
		if(*lpExitCode == STILL_ACTIVE) TerminateProcess(pi.hProcess, NULL);
	}else{
		errorMessage("No active session found!");
	}
}
// Start Signals

void guiSendNetwork(char* buf){
	if(ST_ON){
		int good = sendNetwork(buf);
		if(good == NETWORK_ERROR){
			log("Failed to send command\n", MSG_ERROR);
		}else{
			log("Command '", MSG_NORMAL);
			log(buf, MSG_ATTENTION);
			log("' sent \n", MSG_NORMAL);
		}
	}else{
		errorMessage("No active session found!\nPlease start display grid first");
	}
}


// Image file signals
void onImageFileBrowseClicked() {
	selectFile("image_file_path");
}

void onLoadImageClicked(){	
	int length = strlen(gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(xml, "image_file_path"))));
	FILE* image_file = fopen(gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(xml, "image_file_path"))), "r");
	if(!image_file){
		errorMessage("Can't find image file!\nRemember, Image file has to be in the same location on all clients");
		return;
	}
	fclose(image_file);
    
	if(length > 1){
		char* command = NEW_CHAR((length+8));
		strcpy(command, "7061 ");
		strcpy(&command[5], gtk_entry_get_text(GTK_ENTRY(glade_xml_get_widget(xml, "image_file_path"))));		
		guiSendNetwork(command);
        free(command);
	} else {
		errorMessage("Please select an image file to load");
	}
}
// Image file signals



// Zoom Signals
void onZoomInClicked() {
	guiSendNetwork("5061");
}

void onZoomResetClicked() {
	guiSendNetwork("5063");
}

void onZoomOutClicked() {
	guiSendNetwork("5062");
}
// Zoom Signals

// Full Screen Signals
void setFullScreen(){
	if(ST_ON){
		infoMessage("Full screen applies if set before starting the display!");		
	}
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(xml, "full_screen_checkbutton")))){
		full_screen = 0;
	}else{
		full_screen = 1;
	}
}
// Full Screen Signals

// Pan Signals
void onPanUpClicked() {
	guiSendNetwork("6062");
}

void onPanLeftClicked() {
	guiSendNetwork("6064");
}

void onPanResetClicked() {
	guiSendNetwork("6065");
}

void onPanRightClicked() {
	guiSendNetwork("6063");
}

void onPanDownClicked() {
	guiSendNetwork("6061");
}
// Pan Signals

// Zoom and Pan factors
void onZoomSpinValueChanged(){
    char command[16];    
    sprintf(command, "5071 %3.2f", gtk_spin_button_get_value(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "zoom_spin"))));		
    guiSendNetwork(command);    
}
void onPanSpinValueChanged(){
    char command[16];
    sprintf(command, "5072 %3.2f", gtk_spin_button_get_value(GTK_SPIN_BUTTON(glade_xml_get_widget(xml, "pan_spin"))));		
    guiSendNetwork(command);    
}
// Zoom and Pan factors

DWORD WINAPI NetThread( LPVOID lpParam) {
    //return 0;
    int good;
    clog = fopen("sessionlog.txt", "w");
    if(!clog){
        return 0;
    }
    fprintf(clog, "Logging started at %d\n", time(NULL));
    while(true){
        if(ST_ON){
            good = recvNetwork(net_msgs);
            if(good == NETWORK_OK){
                fprintf(clog, "%d:\t%s\n", time(NULL), net_msgs);
            }
        }
        fflush(clog);
        Sleep(1000);
    }
    fclose(clog);
    return 0; 
}

int main_prog( int argc, char **argv ) {
	gtk_init(&argc, &argv);	
	status = STS_INV;
	xml = glade_xml_new("project2.glade", NULL, NULL); 
	main_window = glade_xml_get_widget(xml, "main_window");

	GtkWidget* text_area = glade_xml_get_widget(xml, "log");
	log_area_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_area)); 
	
	dialog = gtk_file_chooser_dialog_new ("Open File", GTK_WINDOW(main_window),
		GTK_FILE_CHOOSER_ACTION_OPEN,	GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,NULL);
    
	initConfigure();

	net_msgs = NEW_CHAR(NET_BUFFER_SIZE);
	
	// Signals
	SIG_HANDLE("on_quit1_activate",onQuit1Activate);
	SIG_HANDLE("on_help1_activate",onHelp1Activate);
	
	SIG_HANDLE("on_mpi_browse_clicked", onMpiBrowseClicked);
	SIG_HANDLE("on_multdisp_clicked", onMultdispClicked);
	
	SIG_HANDLE("on_start_clicked", onStartClicked);
	SIG_HANDLE("on_stop_clicked", onStopClicked);
	
	SIG_HANDLE("on_image_file_browse_clicked", onImageFileBrowseClicked);
	SIG_HANDLE("on_load_image_clicked", onLoadImageClicked);
	
	SIG_HANDLE("on_zoomin_clicked", onZoomInClicked);
	SIG_HANDLE("on_zoom_reset_clicked", onZoomResetClicked);
	SIG_HANDLE("on_zoomout_clicked", onZoomOutClicked);

	SIG_HANDLE("on_checkbutton1_toggled", setFullScreen);
	
	SIG_HANDLE("on_panup_clicked", onPanUpClicked);
	SIG_HANDLE("on_panleft_clicked", onPanLeftClicked);
	SIG_HANDLE("on_panreset_clicked", onPanResetClicked);
	SIG_HANDLE("on_panright_clicked", onPanRightClicked);
	SIG_HANDLE("on_pandown_clicked", onPanDownClicked);

	SIG_HANDLE("on_window1_delete_event", onQuit1Activate);

    SIG_HANDLE("on_zoom_spin_value_changed", onZoomSpinValueChanged);
    SIG_HANDLE("on_pan_spin_value_changed", onPanSpinValueChanged);

    net_thread = CreateThread(NULL,0,NetThread, NULL, 0, NULL);

	gtk_main();

    fclose(clog);
    TerminateThread(net_thread, NULL);
	return 0;
}

#ifndef _DEBUG
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow){
 	int argc = 1;
	char* argv = "CECA MultiDisplay";
	return main_prog(argc, &argv);
}
#else
int main(int argc, char **argv) {
	return main_prog(argc, argv);
}
#endif


