/*
 * License for Tiled Display modifications
 *
 * Copyright (c) 2005 Mohamed Ibrahim and Stephanie Smullen.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include an acknowlegement that the software contains
 *    code based on contributions made by Mohamed Ibrahim and Stephanie Smullen and
 *    include a link to http://zog.utc.edu/~vislab/.
 *
 * 4.  The Original code contains third party software libraries copyrighted 
 *	under their own licenses. Redistribution of Libraries in source code, 
 *	with or without modification, or redistribution in binary form falls
 *	under the terms of the licenses for the corresponding Libraries. 
 *
 * This software is provided "as is" and any expressed or implied
 * warranties, including, but not limited to, the implied warranties
 * of merchantability and fitness for a particular purpose are
 * disclaimed.  In no event shall the authors or their
 * employers be liable for any direct, indirect, incidental,
 * special, exemplary, or consequential damages (including, but not
 * limited to, procurement of substitute goods or services; loss of
 * use, data, or profits; or business interruption) however caused and
 * on any theory of liability, whether in contract, strict liability,
 * or tort (including negligence or otherwise) arising in any way out
 * of the use of this software, even if advised of the possibility of
 * such damage.
 *

 *
 */
#include "glut.h"
#include "mpi_opengl.h"
#include "file_handler.h"
#include "network_comm.h"
#include "mpi.h"
#include <windows.h>
#include <fcntl.h>		
#include <time.h>
#include <ctype.h>
#include <math.h>

int pixW, pixH;
int leftonly = 1;
//float redcyan = 0;
int stereo = 1;  
int possible = 0;
int mainWindow;
int xsize = 200;
int ysize = 200;
int id_mpi = -1;

// Tiled Display  Data
int success_begin = 0;
double zoom = MIN_ZOOM;
double zoom_factor = 0.25;
double pan[2] = {0.0, 0.0};
double pan_factor = 5.0;
int my_coords[2], grid[2];
int display_id = -1;
char* file_name_local;
int file_name_size;
HANDLE opengl_thread;
HANDLE net_thread;
int glut_argc;
char** glut_argv;
int same_host = 1;
int network_on = 0;
char* net_file;
char* net_msg = (char*)malloc(sizeof(char)*512);

int D_SPACE = 20;
int left_space = D_SPACE;
int right_space = D_SPACE;
int top_space = D_SPACE;
int bottom_space = D_SPACE;

int left_space_enable = 0;
int right_space_enable = 0;
int top_space_enable = 0;
int bottom_space_enable = 0;
// Tiled Display  Data
int xorig=100, yorig=400;
TEXTURE *full;
GLubyte *rightTex; 
FILE* dbg;

/* displays an error message and exits the program */
void die(char *fmt, ...) {
    char* error_msg = (char*)malloc(sizeof(char)*1024);    
    va_list ap;
    va_start(ap, fmt);
    vsprintf(error_msg, fmt, ap);
    fprintf(stderr, error_msg);
    ROOT {
        sendNetwork(error_msg);
        sendNetwork("Exit...");
    }
    free(error_msg);
    end(-1);
}

/* displays adebugging message ifdebugging is turned on */
void debug(char *fmt, ...) {
    char* debug_msg;
    ROOT{
        debug_msg = (char*)malloc(sizeof(char)*1024);    
        va_list ap;
        va_start(ap,fmt);
        vsprintf(debug_msg, fmt, ap);
        sendNetwork(debug_msg);
        free(debug_msg);
        return;
    }
    if (MD_DEBUG){
        debug_msg = (char*)malloc(sizeof(char)*1024);    
        va_list ap;
        va_start(ap,fmt);
        vsprintf(debug_msg, fmt, ap);
        fprintf(MD_DEBUG_FILE, "P[%d]\t", id_mpi);
        fprintf(MD_DEBUG_FILE, debug_msg);
        fflush(MD_DEBUG_FILE);
        free(debug_msg);
    }
}

void doCommand(int id)
{
	debug("doCommand: Entered doCommand, ID = %d\n", id);
	MPI_Bcast(&id, 1, MPI_INT, 0, MPI_COMM_WORLD);
	debug("doCommand:  ID = %d\n", id);
     	
	switch(id){
	case PRT_MSG_RESHAPE_WINDOW: {
		NOTROOT {
			glutReshapeWindow(xsize,ysize);
			glutPositionWindow (xorig, yorig);
		}
		break;
			   }	
	case PRT_MSG_FULL_SCREEN: {
		NOTROOT glutFullScreen(); break;
			   }
	
    case PRT_MSG_INC_HSPACE: {// Horizontal spacing increase
		top_space++; bottom_space++; break;
			   }	
	case PRT_MSG_INC_VSPACE: {// Vertical spacing increase
		left_space++; right_space++; break;
			   }
    case PRT_MSG_DEC_HSPACE: {// Horizontal spacing decrease
		top_space--; bottom_space--; break;
			   }	
	case PRT_MSG_DEC_VSPACE: {// Vertical spacing decrease
		left_space--; right_space--; break;
			   }
        
    case PRT_MSG_ZOOM_IN: {
		zoom += zoom_factor; break;
			   }
	case PRT_MSG_ZOOM_OUT: {
		zoom -= zoom_factor; break;
			   }
	case PRT_MSG_ZOOM_RESET: {
		zoom = 1.0; break;
			   }
    case PRT_MSG_ZOOM_FACTOR_SET: {
        ROOT zoom_factor = atof(net_file)/100.0;
        MPI_Bcast(&zoom_factor, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		break;
			   }
    case PRT_MSG_PAN_FACTOR_SET: {
        ROOT pan_factor = atof(net_file);
        MPI_Bcast(&pan_factor, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		break;
			   }
		
	case PRT_MSG_PAN_UP: {
		pan[1] += pan_factor; break;
			   }
	case PRT_MSG_PAN_DOWN: {
		pan[1] -= pan_factor; break;
			   }
	case PRT_MSG_PAN_LEFT: {
		pan[0] -= pan_factor; break;
			   }
	case PRT_MSG_PAN_RIGHT: {
		pan[0] += pan_factor; break;
			   }
	case PRT_MSG_PAN_RESET: {
		pan[0] = pan[1] = 0.0; break;
	case PRT_MSG_LOAD_FILE:{
		ROOT file_name_size = strlen(net_file)+1;
		MPI_Bcast(&file_name_size, 1, MPI_INT, 0, MPI_COMM_WORLD);
		debug("doCommand: File name size = %ld\n", file_name_size);
		file_name_local = (char*)malloc(sizeof(char)*file_name_size);
		ROOT memcpy(file_name_local, net_file, sizeof(char)*file_name_size);
		file_name_local[file_name_size - 1] = 0;
		MPI_Bcast(file_name_local, file_name_size, MPI_CHAR, 0, MPI_COMM_WORLD);
		debug("doCommand: File name  = %s\n",  file_name_local);
		zoom = MIN_ZOOM;
		pan[0] = pan[1] = 0.0;
		NOTROOT full = read_file(file_name_local);
        NOTROOT prepareImage();
		break;
			  }
			   }
	case PRT_MSG_EXIT: {
		debug("doCommand: Good Bye\n");	
		end(0);
		break;
			   }
	}

	if(zoom > MAX_ZOOM) {
		zoom = MAX_ZOOM;
	}	
	if(zoom < MIN_ZOOM) {
		zoom = MIN_ZOOM;
	}

    if(left_space < 0){
        left_space = 0;
    }
    if(right_space < 0){
        right_space = 0;
    }
    if(top_space < 0){
        top_space = 0;
    }
    if(bottom_space < 0){
        bottom_space = 0;
    }
    
	
	debug("doCommand: Pan %f %f , Zoom %f\n",  pan[0], pan[1], zoom);
	NOTROOT if (rightTex != NULL) {
		glutPostRedisplay();
	}
	
	debug("doCommand: Command #%d received", id);
} 

void prepareImage(void)
{  	
    int i;
    if(full == NULL) return;
	pixW = full->width/2;
	if (leftonly){
		pixW = full->width;
	}
	pixH = full->height;
	
	debug ("prepareImage: pixH = %d pixW = %d :Left\n",pixH, pixW);
	
	
	if (rightTex != NULL){
		free (rightTex);
	}	
	rightTex = (GLubyte *)malloc(pixH*pixW*RGBA);

    for (i=0; i<pixH; i++) {
			memcpy(rightTex + RGBA*(i*pixW), full->tex + RGBA*(i*full->width), RGBA*pixW);
	}
    free(full->tex);
    free(full);
} 

void display(void)
{  
    int wwidth, wheight;
    wwidth = glutGet (GLUT_WINDOW_WIDTH);
	wheight = glutGet (GLUT_WINDOW_HEIGHT);
	glPixelZoom((float)grid[0]*zoom*wwidth/pixW, (float)grid[1]*zoom*wheight/pixH);
	
	glDrawBuffer(GL_BACK);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glRasterPos2d (0,0);
	glBitmap(0,0,0,0,pan[0],pan[1],NULL);
    glDrawPixels(pixW, pixH, GL_RGBA, GL_UNSIGNED_BYTE, rightTex);

	glutSwapBuffers();
} 

void reshape (int w, int h)
{
	xsize = w; ysize = h;
	debug("reshape: My coord [%d,%d], grid [%d,%d], w = %d, h = %d\n", my_coords[0], my_coords[1], grid[0], grid[1], w, h);
	
    NOTROOT glViewport(
        ((-xsize/2) - (my_coords[0]*xsize)) /*- ((left_space_enable) ? left_space : 0 )*/, 
        ((-ysize/2) - (my_coords[1]*ysize)) /*- ((bottom_space_enable) ? bottom_space : 0 )*/, 
        (xsize)                             /*+ ((right_space_enable) ? right_space : 0 )*/, 
        (ysize)                             /*+ ((top_space_enable) ? top_space : 0 )*/);
} 

void initWindow () {
	char* window_title = (char*)malloc(sizeof(char)*64);
	NOTROOT	sprintf(window_title,"Tiled Display [%d,%d]", my_coords[0], my_coords[1]);

	
    if (mainWindow != 0){
		glutDestroyWindow (mainWindow);
	}
    glutInitWindowSize (xsize,ysize); 
	NOTROOT glutInitWindowPosition (xorig, yorig);
	
	
	
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STEREO);
	possible = glutGet(GLUT_DISPLAY_MODE_POSSIBLE);
	
	debug("initWindow: [%d] [%d,%d]\n", id_mpi, xsize, ysize);
	if (possible && stereo) { 
		debug("initWindow: Stereo OK -- set to quad buffer\n");
		mainWindow = glutCreateWindow (window_title);
		debug("initWindow: Should have opened window\n", id_mpi);
	}
	
    else {
		if (!stereo){
			debug ("initWindow: Stereo buffer off -- set to double buffer\n");
		} else {
			debug ("initWindow: No stereo -- set to double buffer\n"); 
		}
		glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
		mainWindow = glutCreateWindow (window_title);
		debug("initWindow: Should have opened window\n", id_mpi);
    }
	
	NOTROOT{ 
		glutDisplayFunc(display); 
		glutReshapeFunc(reshape);
		if(same_host == 0) glutFullScreen();
	}	
}

DWORD WINAPI DisplayThread( LPVOID lpParam) {
	glutInit (&glut_argc, glut_argv);
	initWindow();
	glutMainLoop(); 
	return 0; 
}

void netCommand(char* buf){    
	debug("netCommand: %s\n", buf);
	if(strlen(buf) > 4){ // Assuming an image file name
        net_file = &buf[5];
		debug("net data = '%s'\n", net_file);
        buf[4] = 0;        
		doCommand(atoi(buf));
	}
	else{ // Normal no param command
		doCommand(atoi(buf));
	}
}

// NO DEBUGGING HERE [IT WILL BLOW UP], MULTITHREADING ISSUES
DWORD WINAPI NetThread( LPVOID lpParam) {
    int good = initNetwork(9000,"localhost");
    if(good == NETWORK_OK){
        network_on = 1;
        char* net_recv = (char*)malloc(sizeof(char)*NET_BUFFER_SIZE);
        while(true){ 
            good = recvNetwork(net_recv);
            if(good == NETWORK_OK){
                netCommand(net_recv);
            } else {
                printf("NET Error [Can't talk to server] Exit\n");
                end(1);
                break;
            }
        }
        free(net_recv);
    } else {
        printf("Permenant NET Error [Can't connect to server] Exit\n");
        end(1);
    }
    Sleep(2500);
    return 0; 
}

void end(int returnVal){
	//if(success_begin == 1) {
		TerminateThread(opengl_thread, NULL);
		debug("end: Process %d exiting\n", id_mpi);
		NOTROOT MPI_Barrier(MPI_COMM_WORLD);
	//}
	MPI_Finalize();
	debug("end: Process %d finalized MPI\n", id_mpi);
	if(MD_DEBUG) {
		fclose(dbg);
	}
	exit(returnVal);
}

int main(int argc, char **argv) {
	MPI_Comm cart_comm;
	MPI_Comm display_comm;
	LPDWORD lpExitCode = (LPDWORD)malloc(sizeof(unsigned long));
	int process_count;
	int processes_grid[2];
	int is_periods[2] = {0, 0};
	// Initialize 
	MPI_Init(&argc,&argv);
	glut_argc = argc; glut_argv = argv;
	MPI_Comm_size(MPI_COMM_WORLD, &process_count);
	MPI_Comm_rank(MPI_COMM_WORLD, &id_mpi);
	ROOT net_thread = CreateThread(NULL,0,NetThread, NULL, 0, NULL);
	
	if(argc < 4){
		ROOT fprintf(stderr,"usage:\n%s {grid_width} {grid_height} {one_screen}\n", argv[0]);
		end(1);
	}
	int i;
	for(i = 0; i < argc; i++){
		printf("%d: '%s'\n", i, argv[i]);
	}
	success_begin = 1;
	process_count--;
	grid[0] = atoi(argv[argc - 3]);
	grid[1]= atoi(argv[argc - 2]);;
	same_host = atoi(argv[argc - 1]);
	
	if(grid[0] * grid[1] != process_count){
		ROOT {
			fprintf(stderr,"Number of processes is not accepted %d x %d grid with %d display processes\n", grid[0], grid[1], process_count);
			fprintf(stderr,"argv =  '%s' '%s' \n", argv[1], argv[2]);
		}
		end(-1);
	}
	success_begin = 1;
	// Initialize 
	
	// Separate communicators
	MPI_Comm_split (MPI_COMM_WORLD, (id_mpi == 0), id_mpi, &display_comm );
	// Separate communicators
	char* dbg_name = (char*)malloc(sizeof(char)*32);
	if(MD_DEBUG) {
		sprintf(dbg_name,"DEBUG_%02d.log", id_mpi);
		dbg = fopen(dbg_name,"a"); 
		debug("main: Time:\t%ld\n", time(NULL));
	}		
	// Display Processes	
	NOTROOT {
		// Create cartesian grid
		processes_grid[0] = grid[0];
		processes_grid[1] = grid[1];
		MPI_Cart_create(display_comm, 2, processes_grid, is_periods, false, &cart_comm);
		MPI_Comm_rank(display_comm, &display_id);
		MPI_Cart_coords(cart_comm, display_id, 2, my_coords);
		if(same_host == 1){
			xorig += my_coords[0]*xsize;
			yorig -= my_coords[1]*ysize;
		}
		// Create cartesian grid

        left_space_enable = (my_coords[0] > 0)? 1 : 0;
        right_space_enable = (my_coords[0] < (grid[0]-1))? 1 : 0;
        top_space_enable = (my_coords[1] < (grid[1]-1))? 1 : 0;
        bottom_space_enable = (my_coords[1] > 0)? 1 : 0;

        debug("main: L = %d, R = %d, T = %d, B = %d\n", left_space_enable, right_space_enable, top_space_enable, bottom_space_enable);
		
		// Start Display
		opengl_thread = CreateThread(NULL,0,DisplayThread, NULL, 0, NULL);
		// Start Display
		
		// Display Processes
		
		// Commands
		while(true){
			doCommand(1000000);
		}
	}
	
	MPI_Barrier(MPI_COMM_WORLD);
	// Commands
	
	// End
	
	end(0);
	return 0;
	// End
}
